package org.netkernel.wiki.rm.endpoint;

import java.io.LineNumberReader;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.URI;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;

public class MarkupParserEndpoint extends StandardAccessorImpl
{
	private static final String SUBSTITUTIONKEY="wikiSubstitutionPoint";
	Pattern mMacroPattern=Pattern.compile("\\{([a-zA-Z0-9]*)([\\+]*?.*?)\\}(.*?)\\{/\\1\\}",Pattern.DOTALL|Pattern.MULTILINE);
	Pattern mSubPattern=Pattern.compile(SUBSTITUTIONKEY+"\\d*", Pattern.DOTALL|Pattern.MULTILINE);
	Pattern mActiveArgsPattern=Pattern.compile("\\+([a-zA-Z][0-9a-zA-Z\\-\\._]*)@([^\\+]*)", Pattern.DOTALL);
	
	private static final String WINK_PARSER_BASE="wikiParser";
	private static final String WINK_ENGINE_XHTML_MODE="/XHTML";
	private static final String WINK_ENGINE_DOCBOOK_MODE="/DOCBOOK";
	private static final String WINK_PARSER_XHTML=WINK_PARSER_BASE+WINK_ENGINE_XHTML_MODE;
	
	public MarkupParserEndpoint()
	{	this.declareThreadSafe();
		this.declareSourceRepresentation(String.class);
	}

	@Override
	public void onSource(INKFRequestContext aContext) throws Exception
	{	String type=aContext.getThisRequest().getArgumentValue("activeType");
		boolean xhtml=type.equals(WINK_PARSER_XHTML);
		String cwu=aContext.getCWU();
		if(cwu==null)
		{	aContext.setCWU(aContext.getThisRequest().getArgumentValue("operand"));			
		}
		String opd=(String)aContext.source("arg:operand", String.class);
		LineNumberReader lr=new LineNumberReader(new StringReader(opd));
		String firstline=lr.readLine();
		
		String syntax=null;
		boolean replace=true;
		if(firstline!=null && firstline.startsWith("#"))
		{	syntax=firstline.substring(1);
		}
		else
		{	syntax="MEDIAWIKI";
			replace=false;
		}
		if(replace)
		{	opd=opd.replaceFirst("#.*","");				
		}

		//Deal with macro  templates
		PairList pl=new PairList(4);
		opd=findMacros(opd, pl);
		
		URI base=null;
		if(aContext.getThisRequest().argumentExists("base"))
		{	base=URI.create(aContext.getThisRequest().getArgumentValue("base"));
			if(base.isAbsolute())
			{	base=URI.create(base.getPath());				
			}
		}
		else
		{	base=URI.create("");			
		}
		
		//Request Wiki Parse Engine
		String tempCWU=aContext.getCWU();
		aContext.setCWU(null);
		INKFRequest req=aContext.createRequest("active:wikiParserEngine/"+(xhtml ? "XHTML" : "DOCBOOK")+"/"+syntax);
		req.addArgumentByValue("operand",opd);
		req.addArgument("base", base.toString());
		req.setRepresentationClass(String.class);
		opd=(String)aContext.issueRequest(req);
		aContext.setCWU(tempCWU);

		//Perform macro engine operations
		boolean recurse=true;
		opd=recurseMacros(opd, pl, base.toString(), xhtml, aContext, recurse);
		
		//ByteArrayRepresentation rep=new ByteArrayRepresentation(opd.getBytes());
		INKFResponse resp=aContext.createResponseFrom(opd);
		resp.setMimeType("text/xml");
	}
	
	private String recurseMacros(String source, PairList pl, String base, boolean xhtml,  INKFRequestContext aContext, boolean recurse)
	{	int i=0;
		int start=0;
		Matcher m=mSubPattern.matcher(source);
		int imgidx=1;
		while(m.find(start))
		{	String engineType=(String)pl.getValue1(i);
			MacroContent value=(MacroContent)pl.getValue2(i);
			String intermediateContent=runMacros(engineType, value, base.toString(), xhtml, imgidx, aContext, recurse);
			if(engineType.equals("image"))
			{	imgidx++;
			}
			PairList pl2=new PairList(2);
			intermediateContent=findMacros(intermediateContent, pl2);
			if(recurse)
			{	for(int j=0; j<pl2.size(); j++)
				{	intermediateContent=recurseMacros(intermediateContent, pl2, base, xhtml, aContext, recurse);
				}
			}
			intermediateContent=escapeForMatcherReplace(intermediateContent);
			source=m.replaceFirst(intermediateContent);
			start=m.start();
			m.reset(source);
			i++;
		}
		return source;
	}

	private String runMacros(String engineType, MacroContent aValue, String base, boolean xhtml, int imgidx, INKFRequestContext aContext, boolean recurse)
	{	String value=aValue.content;
		if(engineType.equals("literal"))
		{	recurse=false;
			if(xhtml)
			{	value="<pre class=\"literal\">"+XMLUtils.escape(Matcher.quoteReplacement(value))+"</pre>";
			}
			else
			{	value="<literallayout>"+XMLUtils.escape(Matcher.quoteReplacement(value))+"</literallayout>";
			}
		}
		else
		{	String mode= xhtml ? WINK_ENGINE_XHTML_MODE : WINK_ENGINE_DOCBOOK_MODE;
			try
			{	String cwu=aContext.getCWU();
				if(cwu!=null)
				{	aContext.setCWU(null);
				}
				INKFRequest req=aContext.createRequest("active:wikiTemplateEngine/"+engineType.toUpperCase()+mode);
				req.addArgumentByValue("operand", value);
				if(!"".equals(aValue.argument))
				{	Matcher m=mActiveArgsPattern.matcher(aValue.argument);
					while(m.find())
					{	String n=m.group(1);
						String v=m.group(2);
						req.addArgument(n,v);
					}
				}
				if(engineType.equals("image"))
				{	req.addArgument("index", "i:"+imgidx);
					req.addArgumentByValue("base", base);
					req.addArgument("entry",aContext.getThisRequest().getArgumentValue("operand"));
					recurse=false;
				}
				req.setRepresentationClass(String.class);
				try
				{
					value=(String)aContext.issueRequest(req);
				}
				finally
				{	//Reset cwu
					aContext.setCWU(cwu);
				}
			}
			catch(Exception e)
			{	if(xhtml)
				{	value="<div class=\"error\"><pre><![CDATA["+value+"]]></pre></div>";
				}
				else
				{	//Need correct docbook error mode here
					value="<error>"+XMLUtils.escape(value)+"</error>";
				}
			}
			if(xhtml)
			{	value=Matcher.quoteReplacement(value);
			}
			else
			{	value=Matcher.quoteReplacement(value);					
			}
		}
		return value;
	}
	
	private String findMacros(String source, PairList pl)
	{	Matcher m=mMacroPattern.matcher(source);
		int i=1;
		int start=0;
		while(m.find(start))
		{	pl.put(m.group(1), new MacroContent(m.group(2), m.group(3)));
			source=m.replaceFirst(SUBSTITUTIONKEY+i);
			start=m.start();
			m.reset(source);
			i++;
		}
		return source;
	}
	
	private String escapeForMatcherReplace(String in)
	{	char[] cs=in.toCharArray();
		StringWriter sw=new StringWriter();
		for(int i=0; i<cs.length; i++)
		{	
			char c=cs[i];
			if(cs[i]=='$')
			{	char p=0;
				if(i>0)
				{	p=cs[i-1];				
				}
				if(p!='\\')
				{	sw.append('\\');
				}
			}
			sw.append(c);
		}
		return sw.toString();
	}
	
	class MacroContent
	{
		public String argument;
		public String content;
		
		public MacroContent(String aArgument, String aContent)
		{	argument=aArgument;
			content=aContent;
		}
		
		public String toString()
		{
			return argument+":"+content;
		}
	}
	
}
